/*  multiechoModel.h

    Saad Jbabdi, FMRIB Analysis Group

    Copyright (C) 2008 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#if !defined(multiecho1_h)
#define multiecho1_h

#include "armawrap/newmat.h"
#include "newimage/newimageall.h"
#include "miscmaths/miscmaths.h"
#include "miscmaths/miscprob.h"
#include "miscmaths/nonlin.h"
#include "miscmaths/minimize.h"
#include "utils/options.h"

#include "multiechoOptions.h"

using namespace NEWIMAGE;
using namespace MISCMATHS;
using namespace NEWMAT;


// Model:  S=S0*exp(-TE/T2)
// Params: S0, T2
class SpinEcho {
public:
  SpinEcho():opts(MULTIECHO::multiechoOptions::getInstance()){
    if(opts.verbose.value()) cout << "reading data and brain mask" << endl;
    read_volume(m_mask,opts.maskfile.value());


    read_volume4D(m_data,opts.datafile.value());

    m_te = read_ascii_matrix(opts.tefile.value());

    if(m_te.Nrows()>1) m_te=m_te.t();
    if(m_te.Nrows()>1){
      cerr << "Error: TE file must be either row or column vector" << endl;
      exit(1);
    }

    if(m_te.Ncols() != (int)m_data.tsize()){
      cerr << "Error: data and TE file are not compatible." << endl;
      exit(1);
    }

    m_S0.reinitialize(m_mask.xsize(),m_mask.ysize(),m_mask.zsize());
    copybasicproperties(m_mask,m_S0);
    m_S0=0;

    m_t2.reinitialize(m_mask.xsize(),m_mask.ysize(),m_mask.zsize());
    copybasicproperties(m_mask,m_t2);
    m_t2=0;
    if(opts.savepred.value()){
      m_pred.reinitialize(m_mask.xsize(),m_mask.ysize(),m_mask.zsize(),m_data.tsize());
      copybasicproperties(m_mask,m_pred);
      m_pred=0;
    }

  }
  ~SpinEcho(){}

  void fit();
  void save();

 protected:
  MULTIECHO::multiechoOptions& opts;
  volume4D<float>              m_data;
  volume4D<float>              m_pred;
  Matrix                       m_te;
  volume<float>                m_mask;

  volume<float>                m_t2;
  volume<float>                m_S0;
  Matrix                       m_pinvM;


};



// Model: S = A+B*exp(-C*TI)
// Params: A, B, C=1/T1=exp(Ct)  [positive constraint on C]
class InversionRecovery : public NonlinCF{
public:
  InversionRecovery():opts(MULTIECHO::multiechoOptions::getInstance()){
    if(opts.verbose.value()) cout << "reading data and brain mask" << endl;
    read_volume(m_mask,opts.maskfile.value());


    read_volume4D(m_data,opts.datafile.value());

    m_ti = read_ascii_matrix(opts.tifile.value());

    if(m_ti.Nrows()>1) m_ti=m_ti.t();
    if(m_ti.Nrows()>1){
      cerr << "Error: TI file must be either row or column vector" << endl;
      exit(1);
    }

    if(m_ti.Ncols() != (int)m_data.tsize()){
      cerr << "Error: data and TI file are not compatible." << endl;
      OUT(m_ti.Ncols());
      OUT(m_data.tsize());
      exit(1);
    }

    m_A.reinitialize(m_mask.xsize(),m_mask.ysize(),m_mask.zsize());
    copybasicproperties(m_mask,m_A);
    m_A=0;

    m_B.reinitialize(m_mask.xsize(),m_mask.ysize(),m_mask.zsize());
    copybasicproperties(m_mask,m_B);
    m_B=0;

    m_t1.reinitialize(m_mask.xsize(),m_mask.ysize(),m_mask.zsize());
    copybasicproperties(m_mask,m_t1);
    m_t1=0;

    if(opts.savepred.value()){
      m_pred.reinitialize(m_mask.xsize(),m_mask.ysize(),m_mask.zsize(),m_data.tsize());
      copybasicproperties(m_mask,m_pred);
      m_pred=0;
    }

  }
  ~InversionRecovery(){}


  // routines from NonlinCF
  NEWMAT::ReturnMatrix grad(const NEWMAT::ColumnVector& p)const;
  boost::shared_ptr<BFMatrix> hess(const NEWMAT::ColumnVector&p,boost::shared_ptr<BFMatrix> iptr)const;
  double cf(const NEWMAT::ColumnVector& p)const;
  NEWMAT::ReturnMatrix forwardModel(const NEWMAT::ColumnVector& p)const;



  void fit();
  void save();

 protected:
  MULTIECHO::multiechoOptions& opts;
  volume4D<float>              m_data;
  volume4D<float>              m_pred;
  Matrix                       m_ti;
  volume<float>                m_mask;
  Matrix                       m_pinvM;
  volume<float>                m_t1;
  volume<float>                m_A;
  volume<float>                m_B;
  ColumnVector                 m_Y;


};




#endif
